""" Inter-rater reliability

In statistics, inter-rater reliability (also called by various similar names,
such as inter-rater agreement, inter-rater concordance, inter-observer
reliability, and so on) is the degree of agreement among raters. It is a score
of how much homogeneity or consensus exists in the ratings given by various
judges.

There are a number of statistics that can be used to determine inter-rater
reliability. Different statistics are appropriate for different types of
measurement. Some options are

* joint-probability of agreement
* `Cohen's kappa <https://en.wikipedia.org/wiki/Cohen%27s_kappa>`_
* `Scott's pi <https://en.wikipedia.org/wiki/Scott%27s_pi>`_
* `Fleiss' kappa <https://en.wikipedia.org/wiki/Fleiss%27_kappa>`_
* `Krippendorff's alpha <https://en.wikipedia.org/wiki/Krippendorff%27s_alpha>`_
* `Concordance Correlation Coefficient <https://en.wikipedia.org/wiki/Concordance_correlation_coefficient>`_
* `Intra-Class Correlation <https://en.wikipedia.org/wiki/Intraclass_correlation>`_

and many others.
"""


""" Sample datasets for demonstrating and testing agreement coefficients. """
import pandas as pd


def dist_4cat():
    """ Distribution of 4 raters by Category and Subject - Subjects allocated \
    in 2 groups A and B.
        
    This dataset summarizes the ratings assigned by 4 raters who classified 15
    subjects into one of 3 categories named "a", "b", and "c".

    Returns
    -------
    pandas.DataFrame

        Group
            This variable represents the group name.
        a
            Number of ratings in category "a"
        b
            Number of ratings in category "b"
        c
            Number of ratings in category "c"
    """
    group = pd.Index('A A A A A A A A A B B B B B B'.split(), name='Group')
    data = dict(
        a=[3, 2, 2, 2, 3, 3, 0, 0, 0, 3, 0, 0, 0, 0, 3],
        b=[0, 1, 1, 0, 1, 1, 4, 3, 0, 1, 4, 3, 2, 0, 1],
        c=[1, 1, 1, 2, 0, 0, 0, 1, 4, 0, 0, 1, 2, 4, 0])
    return pd.DataFrame(data, group)


def dist_g1g2():
    """ Distribution of 4 raters by subject and by category, for 14 Subjects \
    that belong to 2 groups "G1" and "G2".

    This dataset contains rating data in the form of a subject-level
    distribution of 4 raters by category the subject was classified into.
    A total of 4 raters had to classify 14 subjects into one of 5 categories
    "a", "b", "c", "d", and "e". None of the 4 raters scored all 14 units.
    Therefore, some missing ratings appear in each of the columns associated
    with the 4 raters.

    This dataset is different version of the more detailed `raw_g1g2` dataset.
    While `raw_g1g2` tells you about the exact category into which each rater
    classified all subjects, this one on the other hand, can only tell you how
    many raters classified a given subject into a particular category.

    See Also
    --------
    datasets.raw_g1g2()

    Returns
    -------
    pandas.DataFrame

        Group
            This variable represents the group name.
        Units
            This variable represents the unit number.
        a
            Number of raters who classified the subject represented by the row
            into category "a"
        b
            Number of raters who classified the subject represented by the row
            into category "b"
        c
            Number of raters who classified the subject represented by the row
            into category "c"
        d
            Number of raters who classified the subject represented by the row
            into category "d"
        e
            Number of raters who classified the subject represented by the row
            into category "e"
    """
    group = 'G2 G1 G2 G1 G1 G1 G2 G1 G2 G2 G1 G1 G2 G2'.split()
    units = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]
    index = pd.MultiIndex.from_arrays([group, units], names=('Group', 'Units'))
    data = dict(
        a=[3, 0, 0, 0, 0, 1, 0, 3, 0, 0, 2, 0, 0, 1],
        b=[0, 3, 0, 0, 4, 1, 0, 1, 4, 0, 0, 0, 3, 2],
        c=[0, 1, 4, 4, 0, 1, 0, 0, 0, 0, 0, 1, 1, 1],
        d=[0, 0, 0, 0, 0, 1, 4, 0, 0, 0, 0, 0, 0, 0],
        e=[0, 0, 0, 0, 0, 0, 0, 0, 0, 3, 0, 0, 0, 0])
    return pd.DataFrame(data, index=index)


def raw_4raters():
    """Rating Data from 4 Raters and 12 Subjects.

    This dataset contains ratings obtained from an experiment where 4 raters
    classified 12 subjects into 5 possible categories labeled as 1, 2, 3, 4,
    and 5. None of the 4 raters scored all 12 units. Therefore, some missing
    ratings in the form of "NA" appear in each of the columns associated with
    the 4 raters. Note that only the 4 last columns are to be used with the
    functions included in this package. The first column only plays a
    descriptive role and is not used in any calculation.
    
    **Source**: Gwet, K.L. (2014) Handbook of Inter-Rater Reliability,
    4th Edition, page #120. Advanced Analytics, LLC

    Returns
    -------
    pandas.DataFrame
        The data frame has the following columns:

        Units
            This variable represents the unit number.
        Rater1
            All ratings from rater 1
        Rater2
            All ratings from rater 2
        Rater3
            All ratings from rater 3
        Rater4
            All ratings from rater 4
    """
    units = pd.Index([1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12], name='Units')
    data = dict(
        Rater1=[1, 2, 3, 3, 2, 1, 4, 1, 2, None, None, None],
        Rater2=[1, 2, 3, 3, 2, 2, 4, 1, 2, 5, None, None],
        Rater3=[None, 3, 3, 3, 2, 3, 4, 2, 2, 5, 1, 3],
        Rater4=[1, 2, 3, 3, 2, 4, 4, 1, 2, 5, 1, None])

    return pd.DataFrame(data, index=units)


def raw_5observers():
    """ Scores assigned by 5 observers to 20 experimental units.

    This dataset contains data from a reliability experiment where 5 observers
    scored 15 units on a 4-point numeric scale based on the values
    0, 1, 2 and 3.

    The dataset has 15 rows (for the 15 subjects) and 6 columns.
    Only the last 5 columns associated with the 5 observers are used in the
    calculations. Of the 5 observers, only observer 3 scored all 15 units.
    Therefore, some missing ratings in the form of "NA" appear in the columns
    associated with the remaining 4 observers.

    **Source**: Gwet, K.L. (2014) Handbook of Inter-Rater Reliability,
    4th Edition. Advanced Analytics, LLC.
    A larger version of this table can be found on page #125

    Returns
    -------
    pandas.DataFrame
        The data frame has the following columns:

        Units
            This variable represents the unit number.
        Observer1
            All ratings from observer 1
        Observer2
            All ratings from observer 2
        Observer3
            All ratings from observer 3
        Observer4
            All ratings from observer 4
        Observer5
            All ratings from observer 5
            
    """
    units = pd.Index(
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15], name='Units')
    data = dict(
        Observer1=[
            1.0, 1.0, 2.0, None, 0.0, 0.0, 1.0, 1.0, 2.0, 2.0, None, 0.0, 1.0,
            3.0, 1.0
        ],
        Observer2=[
            1.0, 1.0, 3.0, 0.0, 0.0, 0.0, 0.0, None, 2.0, 1.0, 1.0, 0.0, 2.0,
            3.0, 1.0
        ],
        Observer3=[2, 0, 3, 0, 0, 0, 2, 2, 2, 1, 0, 0, 2, 2, 1],
        Observer4=[
            None, 1.0, 3.0, None, None, None, None, 0.0, None, 1.0, 0.0, 0.0,
            2.0, 2.0, None
        ],
        Observer5=[
            2.0, None, None, 0.0, 0.0, 0.0, 1.0, None, 2.0, None, None, None,
            None, 3.0, 1.0
        ])
    return pd.DataFrame(data, index=units)


def raw_ben_gerry():
    """Ratings of 12 units from 2 raters named Ben and Gerry.

    This dataset contains ratings that 2 raters named Ben and Gerry assigned to
    12 units distributed in 2 groups "G1" and "G2". Each row of this dataset
    describes an interval and the interpretation of the magnitude it represents.

    Returns
    -------
    pandas.Dataframe
        The data frame has the following columns:

        Group
            This variable represents the group membership.
        Units
            This variable represents the unit number.
        Ben
            All ratings from Ben
        Gerry
            All ratings from Gerry
    """
    group = 'G2 G2 G1 G1 G1 G2 G1 G2 G2 G1 G1 G2'.split()
    units = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12]
    index = pd.MultiIndex.from_arrays(
        [group, units], names=('Group', 'Units'))
    data = dict(
        Ben=['a', 'a', 'b', 'b', 'd', 'c', 'c', 'c', 'e', 'd', 'd', 'a'],
        Gerry=['b', 'a', 'b', 'b', 'b', 'c', 'c', 'c', 'e', None, 'd', 'd'])
    return pd.DataFrame(data, index=index)


def raw_g1g2():
    """ Dataset of raw ratings from 4 Raters on 14 Subjects that belong to 2 \
    groups "G1" and "G2".

    This dataset contains data from a reliability experiment where 4 raters
    identified as Rater1, Rater2, Rater3 and Rater4 scored 14 units on a
    5-point alphabetical scale based on the values a, b, c, d and e.
    These 14 units are allocated to 2 groups named G1 and G2

    This dataset contains ratings obtained from an experiment where 4 raters
    classified 14 subjects into 5 possible categories labeled as a, b, c, d,
    and e. None of the 4 raters scored all 14 units. Therefore, some missing
    ratings appear in each of the columns associated with the 4 raters.

    Returns
    -------
    pandas.DataFrame
        The data frame has the following columns:

        Group
            This variable represents the group membership.
        Units
            This variable represents the unit number.
        Rater1
            All ratings from rater 1
        Rater2
            All ratings from rater 2
        Rater3
            All ratings from rater 3
        Rater4
            All ratings from rater 4
    """
    group = 'G2 G1 G2 G1 G1 G1 G2 G1 G2 G2 G1 G1 G2 G2'.split()
    units = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14]
    index = pd.MultiIndex.from_arrays(
        [group, units], names=('Group', 'Units'))
    data = dict(
        Rater1=[
            'a', 'b', 'c', 'c', 'b', 'a', 'd', 'a', 'b', None, None, None, 'b',
            'b'
        ],
        Rater2=[
            'a', 'b', 'c', 'c', 'b', 'b', 'd', 'a', 'b', 'e', None, None, 'b',
            'a'
        ],
        Rater3=[
            None, 'c', 'c', 'c', 'b', 'c', 'd', 'b', 'b', 'e', 'a', 'c', 'c',
            'c'
        ],
        Rater4=[
            'a', 'b', 'c', 'c', 'b', 'd', 'd', 'a', 'b', 'e', 'a', None, 'b',
            'b'
        ])
    return pd.DataFrame(data, index=index)


def raw_gender():
    """ Rating Data from 4 Raters and 15 human Subjects, 9 of whom are female \
    and 6 males.

    This dataset contains data from a reliability experiment where 4 raters
    scored 15 units on a 3-point alphabetic scale based on the values
    a, b, and c.

    Returns
    -------
    pandas.DataFrame
        The data frame has the following columns:

        Group
            This variable represents the group membership.
        Units
            This variable represents the unit number.
        Rater1
            All ratings from rater 1
        Rater2
            All ratings from rater 2
        Rater3
            All ratings from rater 3
        Rater4
            All ratings from rater 4
    """
    group = [
        'MALE', 'MALE', 'MALE', 'MALE', 'MALE', 'MALE', 'MALE', 'MALE', 'MALE',
        'FEMALE', 'FEMALE', 'FEMALE', 'FEMALE', 'FEMALE', 'FEMALE'
    ]
    units = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15]
    index = pd.MultiIndex.from_arrays(
        [group, units], names=('Group', 'Units'))
    data = dict(
        Rater1=[
            'a', 'a', 'a', 'a', 'a', 'b', 'b', 'b', 'c', 'a', 'a', 'a', 'a',
            'a', 'b'
        ],
        Rater2=[
            'a', 'a', 'a', 'a', 'b', 'a', 'b', 'c', 'c', 'a', 'a', 'a', 'a',
            'b', 'a'
        ],
        Rater3=[
            'a', 'b', 'b', 'c', 'a', 'a', 'b', 'b', 'c', 'a', 'b', 'b', 'c',
            'a', 'a'
        ],
        Rater4=[
            'c', 'c', 'c', 'c', 'a', 'a', 'b', 'b', 'c', 'c', 'c', 'c', 'c',
            'a', 'a'
        ])
    return pd.DataFrame(data, index=index)


def table_cont3x3abstractors():
    """ Distribution of 100 pregnant women by pregnancy type and by abstractor.

    This dataset contains pregnancy type data collected from 100 women who
    entered an Emergency Room with a positive pregnancy test and a second
    condition, which is either abdominal pain or vaginal bleeding. After
    reviewing their medical records, 2 reviewers (also referred to as
    abstractors) classified them into one of the following three pregnancy
    categories: Ectopic Pregnancy (Ectopic), Abnormal Intrauterine pregnancy
    (AIU) and Normal Intrauterine Pregnancy (NIU).

    Each row of this dataset describes an interval and the interpretation of
    the magnitude it represents.

    **Source:** Gwet, K.L. (2014). Handbook of Inter-Rater Reliability,
    4th Edition. Advanced Analytics, LLC.

    Returns
    -------
    pandas.DataFrame
    
        Ectopic
            Ectopic Pregnancy
        AIU
            Abnormal Intrauterine Pregnancy
        NIU
            Normal Intrauterine Pregnancy

    """
    data = {'Ectopic': [13, 0, 0], 'AIU': [0, 20, 4], 'NIU': [0, 7, 56]}
    return pd.DataFrame(data, index=data.keys())


def table_cont4x4diagnosis():
    """ Distribution of 223 Psychiatric Patients by Type of Psychiatric \
    Disorder and Diagnosis Method.

    This dataset shows the distribution of 223 psychiatric patients by
    diagnosis category and by the method used to obtain the diagnosis in a 4x4
    squared table. The first method named "Clinical Diagnosis" (also known as
    "Facility Diagnosis") is used in a service facility (e.g. public hospital,
    or a community unit) and does not rely on a rigorous application of
    research criteria. The second method known as "Research Diagnosis" is based
    on a strict application of research criteria. Column 1 contains the
    diagnosis categories into which patients are classified with Method 1.
    The first row on the other hand, shows categories into which patients are
    classified with Method 2.

    Returns
    -------
    pandas.DataFrame

        Method
            The method used for the diagnosis.
        Diagnosis
            The category of the diagnosis.
    """
    methods = ['Clinical Diagnosis', 'Research Diagnosis']
    diagnosis = ['Schizophrenia', 'Bipolar Disorder', 'Depression', 'Other']
    cols = pd.MultiIndex.from_product(
        [[methods[0]], diagnosis], names=['Method', 'Diagnosis'])
    idx = pd.MultiIndex.from_product(
        [[methods[1]], diagnosis], names=['Method', 'Diagnosis'])
    data = [[40, 4, 4, 17], [6, 25, 2, 13], [4, 1, 21, 12], [15, 5, 9, 45]]
    return pd.DataFrame(data, index=diagnosis, columns=diagnosis)


def distrib_6raters():
    """ Distribution of 6 psychiatrists by Subject/patient and diagnosis \
    category.

    This dataset summarizes the ratings assigned by 6 psychiatrists classifying
    15 patients into one of five categories named "Depression", "Personal
    Disorder", "Schizophrenia", "Neurosis" and "Other.

    **Source:** Fleiss, J. L. (1971). Measuring nominal scale agreement among
    many raters, Psychological Bulletin, 76, 378-382.

    Returns
    -------
    pandas.DataFrame

        Units
            This variable represents the unit number.
        Depression
            The number of raters assigned the subject to the depression category
        Personality Disorder
            The number of raters assigned the subject to the personality
            disorder category
        Schizophrenia
            The number of raters assigned the subject to the schizophrenia
            category
        Neurosis
            The number of raters assigned the subject to the neurosis category
        Other
            The number of raters assigned the subject to the other category
    """
    units = pd.Index(
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15], name='Units')
    data = {
        'Depression': [0, 0, 0, 0, 0, 2, 0, 2, 2, 0, 1, 1, 0, 1, 0],
        'Personality Disorder': [0, 3, 1, 0, 3, 0, 0, 0, 0, 0, 0, 1, 3, 0, 2],
        'Schizophrenia': [0, 0, 4, 0, 0, 4, 4, 3, 0, 0, 0, 0, 3, 0, 0],
        'Neurosis': [6, 0, 0, 0, 3, 0, 0, 1, 4, 0, 5, 4, 0, 5, 3],
        'Other': [0, 3, 1, 6, 0, 0, 2, 0, 0, 6, 0, 0, 0, 0, 1]
    }
    return pd.DataFrame(data, index=units)



""" Chance-corrected Agreement Coefficient for tabular ratings.

Examples
--------
>>> from irrCAC.datasets import table_cont3x3abstractors
>>> from irrCAC.table import CAC
>>> data = table_cont3x3abstractors()  # doctest: +NORMALIZE_WHITESPACE
>>> print(data)
         Ectopic  AIU  NIU
Ectopic       13    0    0
AIU            0   20    7
NIU            0    4   56

Initialize a CAC object with the data from the contigency table

>>> cont3x3abstractors = CAC(data)
>>> print(cont3x3abstractors.gwet())  # doctest: +NORMALIZE_WHITESPACE
'est': {'coefficient_value': 0.84933,
        'coefficient_name': "Gwet's AC1",
        'confidence_interval': (0.76358, 0.93508),
        'p_value': 0.0,
        'z': 19.65248,
        'se': 0.04322,
        'pa': 0.89,
        'pe': 0.26992},
'weights': array([[1., 0., 0.],
                  [0., 1., 0.],
                  [0., 0., 1.]]),
'categories': ['Ectopic', 'AIU', 'NIU']}

To use weights with the calculations, we pass the type of weights as argument.

>>> cont3x3abstractors_quadratic = CAC(data, weights='quadratic')
>>> print(cont3x3abstractors_quadratic.gwet())  # doctest: +NORMALIZE_WHITESPACE
{'est': {'coefficient_value': 0.94024,
         'coefficient_name': "Gwet's AC2",
         'confidence_interval': (0.90468, 0.97579),
         'p_value': 0.0,
         'z': 52.46802,
         'se': 0.01792,
         'pa': 0.9725,
         'pe': 0.53985},
'weights': array([[1.  , 0.75, 0.  ],
                  [0.75, 1.  , 0.75],
                  [0.  , 0.75, 1.  ]]),
'categories': ['Ectopic', 'AIU', 'NIU']}
"""
import numpy as np
from scipy import stats
from irrCAC.weights import Weights


class CAC:
    """ Chance-corrected Agreement Coefficients (CAC) 
    
    The following chance-corrected agreement coefficients (CAC) among 2 raters
    are provided.

    * Brennan-Prediger,
    * Cohen's kappa,
    * Gwet's AC1/AC2,
    * Krippendorff's Alpha,
    * Percent Agreement,
    * Scott's Pi.

    Multiple sets of weights are proposed for computing weighted analyses.
    All of these statistical procedures are described in details in Gwet, K.L.
    (2014,ISBN:978-0970806284): "Handbook of Inter-Rater Reliability,"
    4th edition, Advanced Analytics, LLC.

    Parameters
    ----------
    ratings : DataFrame
        A data frame of ratings where each column represents one rater and
        each row one subject.
    weights : array-like, ndarray, or str, {"identity", "quadratic", "ordinal",\
    "linear", "radical", "ratio", "circular", "bipolar"}
        A mandatory parameter that is either a string variable or a matrix.
        The string describes one of the predefined weights. If this
        parameter is a matrix then it must be a square matrix qxq where q
        is the number of possible categories where a subject can be
        classified. If some of the q possible categories are not used,
        then it is strongly advised to specify the complete list of
        possible categories as a vector in parameter ``categories``.
        Otherwise, the program may not work.
    confidence_level : float, default 0.95
        An optional parameter representing the confidence level associated
        with the confidence interval. Its default value is 0.95.
    N : int, default infinity
        An optional parameter representing the population size (if any).
        It may be used to perform the final population correction to the
        variance. Its default value is infinity.
    digits : int, default 5
        The number of digits to round the results.
    """
    def __init__(
            self,
            ratings,
            weights='identity',
            confidence_level=0.95,
            N=np.inf,
            digits=5):
        weights_choices = (
            "identity", "quadratic", "ordinal", "linear",
            "radical", "ratio", "circular", "bipolar")
        if weights not in weights_choices:
            raise ValueError(f'weights values can be any of {weights_choices}')
        assert 0 < confidence_level <= 0.99,\
            'Confidence level should be in the range (0, 1).'
        self.confidence_level = confidence_level

        if isinstance(weights, str):
            self.weights_name = weights
            weights_functions = Weights(list(range(1, len(ratings) + 1)))
            self.weights_mat = weights_functions[self.weights_name]
        else:
            self.weights_name = 'Custom Weights'
            self.weights_mat = np.asarray(weights)
            if ratings.shape[0] != self.weights_mat.shape[0]:
                raise AttributeError(
                    f'We expect a {ratings.shape[0]}x{ratings.shape[1]}'
                    f' weights matrix but the input matrix is '
                    f'{self.weights_mat.shape[0]}x{self.weights_mat.shape[1]}.'
                )

        if ratings.shape[0] != ratings.shape[1]:
            raise AttributeError(
                'The contingency table should have the same '
                'number of rows and columns.')
        self.ratings = ratings
        self.n = np.sum(ratings.values)
        self.f = self.n / N
        self.q = len(self.ratings)
        self.pa = np.sum(self.ratings.values * self.weights_mat / self.n)
        self.digits = digits
        self.agreement = {
            'est': {
                'coefficient_value': 0,
                'coefficient_name': None,
                'confidence_interval': (0, 0),
                'p_value': 0,
                'z': 0,
                'se': 0,
                'pa': self.pa,
                'pe': 0,
            },
            'weights': self.weights_mat,
            'categories': self.ratings.index.to_list()
        }

    def __str__(self):
        subjects = f'Subjects: {self.n}'
        categories = f'Categories: {self.agreement["categories"]}'
        weights_name = f'Weights: "{self.weights_name}"'
        class_path = f'{CAC.__module__}.{CAC.__name__}'
        _str = f'{class_path} {subjects}, {categories}, {weights_name}'
        return f'<{_str}>'

    def __repr__(self):
        return self.__str__()

    def bp(self):
        """ Brennan-Prediger coefficient for 2 raters.
        """

        tw = np.sum(self.weights_mat)
        pe = tw / pow(self.q, 2)
        bp_coeff = (self.pa - pe) / (1 - pe)
        pkl = self.ratings.values / self.n
        sum1 = 0
        for k in range(self.q):
            for l in range(self.q):
                sum1 += pkl[k, l] * self.weights_mat[k, l]**2
        var_bp = ((1 - self.f) / (self.n * (1 - pe)**2)) * (sum1 - self.pa**2)
        stderr = np.sqrt(var_bp)
        p_value = 2 * (1 - stats.t.cdf(abs(bp_coeff / stderr), self.n - 1))
        lcb, ucb = stats.t.interval(
            alpha=self.confidence_level,
            df=self.n - 1,
            scale=stderr,
            loc=bp_coeff)
        ucb = min(1, ucb)
        self.agreement['est'].update(
            dict(
                coefficient_name='Brennan-Prediger',
                pa=np.round(self.pa, self.digits),
                pe=round(pe, self.digits),
                se=round(stderr, self.digits),
                z=round(bp_coeff / stderr, self.digits),
                coefficient_value=round(bp_coeff, self.digits),
                confidence_interval=(
                    round(lcb, self.digits), round(ucb, self.digits)),
                p_value=round(p_value, self.digits)))
        return self.agreement

    def cohen(self):
        """ Cohen's kappa coefficient for 2 raters.
        
        Cohen's kappa measures the agreement between two raters who each
        classify N subjects into :math:`q` mutually exclusive categories.
        """
        pk_dot = (self.ratings.sum(axis=1) / self.n).values.reshape(-1, 1)
        p_dot_l = (self.ratings.sum(axis=0) / self.n).values.reshape(-1, 1)
        pe = np.sum(self.weights_mat * np.matmul(pk_dot, p_dot_l.T))
        kappa = (self.pa - pe) / (1 - pe)
        pkl = self.ratings.values / self.n
        pb_dot_k = (self.weights_mat * p_dot_l).sum(axis=0)
        pbl_dot = (self.weights_mat * pk_dot).sum(axis=0)
        sum1 = 0
        for k in range(self.q):
            for l in range(self.q):
                sum1 += pkl[k][l] * (
                    self.weights_mat[k][l] - (1 - kappa) *
                    (pb_dot_k[k] + pbl_dot[l]))**2
        var_kappa = ((1 - self.f) / (self.n * (1 - pe)**2)) * (
            sum1 - (self.pa - 2 * (1 - kappa) * pe)**2)
        stderr = np.sqrt(var_kappa)
        p_value = 2 * (1 - stats.t.cdf(abs(kappa / stderr), self.n - 1))
        lcb, ucb = stats.t.interval(
            alpha=self.confidence_level,
            df=self.n - 1,
            scale=stderr,
            loc=kappa)
        ucb = min(1, ucb)
        self.agreement['est'].update(
            dict(
                coefficient_name="Cohen's kappa",
                pa=np.round(self.pa, self.digits),
                pe=np.round(pe, self.digits),
                se=np.round(stderr, self.digits),
                z=np.round(kappa / stderr, self.digits),
                coefficient_value=np.round(kappa, self.digits),
                confidence_interval=(
                    np.round(lcb, self.digits), np.round(ucb, self.digits)),
                p_value=np.round(p_value, self.digits)))
        return self.agreement

    def gwet(self):
        """ Gwet's AC1/AC2 coefficient for 2 raters.

        The AC1 coefficient was suggested by Gwet (2008a) as a paradox-resistant
        alternative to Cohen’s Kappa. The percent chance agreement it is
        defined as the propensity for raters to agree on hard-to-score
        subjects and is calculated by multiplying the probability to agree
        when the rating is random by the probability to select a
        hard-to-score subject.

        The Gwet's AC2 coefficient is the one when using weight for the
        calculation.
        """
        pk_dot = (self.ratings.sum(axis=1) / self.n).values.reshape(-1, 1)
        p_dot_l = (self.ratings.sum(axis=0) / self.n).values.reshape(-1, 1)
        pi_dot_k = (pk_dot + p_dot_l) / 2
        tw = np.sum(self.weights_mat)
        pe = tw * np.sum(pi_dot_k * (1 - pi_dot_k)) / (self.q * (self.q - 1))
        ac1 = (self.pa - pe) / (1 - pe)
        pkl = self.ratings.values / self.n
        sum1 = 0
        for k in range(self.q):
            for l in range(self.q):
                sum1 += float(
                    pkl[k][l] * (
                        self.weights_mat[k][l] - 2 * (1 - ac1) * tw *
                        (1 - (pi_dot_k[k] + pi_dot_k[l]) / 2) /
                        (self.q * (self.q - 1)))**2)
        var_gwet = ((1 - self.f) / (self.n * (1 - pe)**2)) * (
            sum1 - (self.pa - 2 * (1 - ac1) * pe)**2)
        stderr = np.sqrt(var_gwet)
        p_value = 2 * (1 - stats.t.cdf(abs(ac1 / stderr), self.n - 1))
        lcb, ucb = stats.t.interval(
            alpha=self.confidence_level, df=self.n - 1, scale=stderr, loc=ac1)
        ucb = min(1, ucb)
        if np.sum(self.weights_mat) == self.q:
            coeff_name = "Gwet's AC1"
        else:
            coeff_name = "Gwet's AC2"
        self.agreement['est'].update(
            dict(
                coefficient_name=coeff_name,
                pa=np.round(self.pa, self.digits),
                pe=np.round(pe, self.digits),
                se=np.round(stderr, self.digits),
                z=np.round(ac1 / stderr, self.digits),
                coefficient_value=np.round(ac1, self.digits),
                confidence_interval=(
                    np.round(lcb, self.digits), np.round(ucb, self.digits)),
                p_value=np.round(p_value, self.digits)))
        return self.agreement

    def krippendorff(self):
        """ Krippendorff’s Alpha coefficient for 2 raters.

        .. versionadded:: 0.2.0
        """
        epsi = 1 / (2 * self.n)
        pa = (1 - epsi) * self.pa + epsi
        pk_dot = (self.ratings.sum(axis=1) / self.n).values.reshape(-1, 1)
        p_dot_l = (self.ratings.sum(axis=0) / self.n).values.reshape(-1, 1)
        pi_dot_k = (pk_dot + p_dot_l) / 2
        pe = np.sum(self.weights_mat * (pi_dot_k * pi_dot_k.T))
        kripen_coeff = (pa - pe) / (1 - pe)
        pkl = self.ratings.values / self.n
        pb_dot_k = (self.weights_mat * p_dot_l).sum(axis=0)
        pbl_dot = (self.weights_mat * pk_dot).sum(axis=0)
        pbk = (pb_dot_k + pbl_dot) / 2
        kcoeff = (self.pa - pe) / (1 - pe)
        sum1 = 0
        for k in range(self.q):
            for l in range(self.q):
                sum1 += pkl[k][l] * (
                        self.weights_mat[k][l] - (1 - kcoeff) *
                        (pbk[k] + pbk[l]))**2
        var_kripp = ((1 - self.f) / (self.n * (1 - pe)**2)) * (
                sum1 - (self.pa - 2 * (1 - kcoeff) * pe)**2)
        stderr = np.sqrt(var_kripp)
        p_value = 2 * (1 - stats.t.cdf(abs(kripen_coeff / stderr), self.n - 1))
        lcb, ucb = stats.t.interval(
            alpha=self.confidence_level,
            df=self.n - 1,
            scale=stderr,
            loc=kripen_coeff)
        ucb = min(1, ucb)
        self.agreement['est'].update(
            dict(
                coefficient_name="Krippendorff's Alpha",
                pa=np.round(pa, self.digits),
                pe=np.round(pe, self.digits),
                se=np.round(stderr, self.digits),
                z=np.round(kripen_coeff / stderr, self.digits),
                coefficient_value=np.round(kripen_coeff, self.digits),
                confidence_interval=(
                    np.round(lcb, self.digits), np.round(ucb, self.digits)),
                p_value=np.round(p_value, self.digits)))
        return self.agreement

    def pa2(self):
        """ Percent Agreement coefficient for 2 raters.

        .. versionadded:: 0.2.0
        """
        pkl = self.ratings.values / self.n
        sum1 = 0
        for k in range(self.q):
            for l in range(self.q):
                sum1 += pkl[k][l] * self.weights_mat[k][l] ** 2
        var_pa = ((1  -  self.f) / self.n) * (sum1 - self.pa ** 2)
        stderr = np.sqrt(var_pa)
        p_value = 2 * (1 - stats.t.cdf(abs(self.pa / stderr), self.n - 1))
        lcb, ucb = stats.t.interval(
            alpha=self.confidence_level,
            df=self.n - 1,
            scale=stderr,
            loc=self.pa)
        ucb = min(1, ucb)
        self.agreement['est'].update(
            dict(
                coefficient_name="Percent Agreement",
                pa=np.round(self.pa, self.digits),
                pe=0,
                se=np.round(stderr, self.digits),
                z=np.round(self.pa / stderr, self.digits),
                coefficient_value=np.round(self.pa, self.digits),
                confidence_interval=(
                    np.round(lcb, self.digits), np.round(ucb, self.digits)),
                p_value=np.round(p_value, self.digits)))
        return self.agreement

    def scott(self):
        """ Scott’s Pi coefficient for 2 raters.
         
         .. versionadded:: 0.2.0
         """
        pk_dot = (self.ratings.sum(axis=1) / self.n).values.reshape(-1, 1)
        p_dot_l = (self.ratings.sum(axis=0) / self.n).values.reshape(-1, 1)
        pi_dot_k = (pk_dot + p_dot_l) / 2
        pe = np.sum(self.weights_mat * (pi_dot_k * pi_dot_k.T))
        scott = (self.pa - pe) / (1 - pe)
        pkl = self.ratings.values / self.n
        pb_dot_k = (self.weights_mat * p_dot_l).sum(axis=0)
        pbl_dot = (self.weights_mat * pk_dot).sum(axis=0)
        pbk = (pb_dot_k + pbl_dot) / 2
        sum1 = 0
        for k in range(self.q):
            for l in range(self.q):
                sum1 += pkl[k][l] * (
                        self.weights_mat[k][l] - (1 - scott) *
                        (pbk[k] + pbk[l]))**2
        var_scott = ((1 - self.f) / (self.n * (1 - pe)**2)) * (
                sum1 - (self.pa - 2 * (1 - scott) * pe)**2)
        stderr = np.sqrt(var_scott)
        p_value = 2 * (1 - stats.t.cdf(abs(self.pa / stderr), self.n - 1))
        lcb, ucb = stats.t.interval(
            alpha=self.confidence_level,
            df=self.n - 1,
            scale=stderr,
            loc=scott)
        ucb = min(1, ucb)
        self.agreement['est'].update(
            dict(
                coefficient_name="Scott's Pi",
                pa=np.round(self.pa, self.digits),
                pe=0,
                se=np.round(stderr, self.digits),
                z=np.round(scott / stderr, self.digits),
                coefficient_value=np.round(scott, self.digits),
                confidence_interval=(
                    np.round(lcb, self.digits), np.round(ucb, self.digits)),
                p_value=np.round(p_value, self.digits)))
        return self.agreement



""" A set of predefined weight schemes that can be used with the agreement \
coefficients.
"""
import numpy as np
import pandas as pd


class Weights:
    """ Methods for computing weights for a set of categories.

    The class can compute weights, using the methods:

    * bipolar
    * circular
    * identity
    * linear
    * ordinal
    * quadratic
    * radial
    * ratio
    """
    def __init__(self, categories):
        """ Initialize a `Weights` class based on the rating categories.

        Parameters
        ----------
        categories : list, array-like or data frame
            A mandatory parameter representing the vector of all possible
            ratings. If the `categories` has numerical values, the vector is
            sorted. If the `categories` are categorical labels, the vector is
            a sequence of numbers in range `1…len(categories)`.

        Raises
        ------
        ValueError

            * The provided value for `categories` is not one of a list, a numpy
              array or a pandas dataframe.
            * Giving an unknown name of weights type.
        """
        if isinstance(categories, list):
            self.q = len(categories)
        elif isinstance(categories, (np.ndarray, pd.DataFrame)):
            self.q = categories.shape[-1]
        else:
            raise ValueError(
                'Valid input for `categories` is one of '
                'list, numpy array, or pandas data frame.')
        if all(isinstance(n, (int, float)) for n in categories):
            self.categ_vec = sorted(categories)
        else:
            self.categ_vec = list(range(1, len(categories) + 1))
        self.xmin, self.xmax = min(self.categ_vec), max(self.categ_vec)

    def __getitem__(self, item):
        if item == 'bipolar':
            return self.bipolar()
        elif item == 'circular':
            return self.circular()
        elif item == 'identity':
            return self.identity()
        elif item == 'linear':
            return self.linear()
        elif item == 'ordinal':
            return self.ordinal()
        elif item == 'quadratic':
            return self.quadratic()
        elif item == 'radical':
            return self.radical()
        elif item == 'ratio':
            return self.ratio()
        else:
            raise ValueError(f'"{item} is an unknown type of weights.')

    def __str__(self):
        return f'Weights for {self.q} categories.'

    def bipolar(self):
        """ Function for computing the Bipolar Weights

        Bipolar weights of a matrix :math:`W \in R^{q \\times q}`, where
        :math:`q` is the number of `categories`, are defined for each cell
        :math:`w_{kl}, (k,l=1, \ldots, q)` by

        .. math::
            w_{kl} = \\frac{(k - l)^2}{
                ((k + l - 2 \cdot \min(c))(2 \cdot \max(c) - k - l))}

        where :math:`c \in R^q` is the vector of the `categories`.

        Returns
        -------
        2D :math:`q \\times q` matrix
            A square matrix of bipolar weights to be used for calculating the
            weighted coefficients.
        """
        weights = np.eye(self.q)
        for k in range(self.q):
            for l in range(self.q):
                if k != l:
                    weights[k][l] = pow(
                        self.categ_vec[k] - self.categ_vec[l], 2) / (
                            (
                                self.categ_vec[k] + self.categ_vec[l] -
                                2 * self.xmin) * (
                                    2 * self.xmax - self.categ_vec[k] -
                                    self.categ_vec[l]))
                else:
                    weights[k][l] = 0
        weights = 1 - weights / np.max(weights)
        return weights

    def circular(self):
        """ Function for computing the Circular Weights

        Circular weights of a matrix :math:`W \in R^{q \\times q}`, where
        :math:`q` is the number of `categories`, are defined for each cell
        :math:`w_{kl}, (k,l=1, \ldots, q)` by

        .. math::
            w_{kl} = \sin(\\frac{\pi (k - l)}{(\max(c) - \min(c) + 1)})^2

        where :math:`c \in R^q` is the vector of the `categories`.

        Returns
        -------
        2D :math:`q \\times q` matrix
            A square matrix of circular weights to be used for calculating the
            weighted coefficients.
        """
        weights = np.eye(self.q)
        U = self.xmax - self.xmin + 1
        for k in range(self.q):
            for l in range(self.q):
                weights[k][l] = pow(
                    np.sin(
                        np.pi * (self.categ_vec[k] - self.categ_vec[l]) / U),
                    2)
        weights = 1 - weights / np.max(weights)
        return weights

    def identity(self):
        """ Function for computing the Identity Weights.

        The identity weighted matrix :math:`W \in R^{q \\times q}`, where
        :math:`q` is the number of `categories`, is the same as to calculate the
        coefficients without weights (unweighted). The weights are defined as
        :math:`w_{kk}=1, (k=1, \ldots, q)` and
        :math:`w_{kl}=0, (k,l=1, \ldots, q)` if :math:`k \\neq l`.

        Returns
        -------
        2D :math:`q \\times q` matrix
            A square matrix of identity weights to be used for calculating the
            weighted coefficients.
        """
        weights = np.eye(self.q)
        return weights

    def linear(self):
        """ Function for computing the Linear Weights.

        Linear weights of a matrix :math:`W \in R^{q \\times q}`, where
        math:`q` is the number of `categories`, are defined for each cell
        :math:`w_{kl}, (k,l=1, \ldots, q)` by

        .. math::
            w_{kl} = 1 - \\frac{|k - l|}{q - 1}

        Returns
        -------
        2D :math:`q \\times q` matrix
            A square matrix of linear weights to be used for calculating the
            weighted coefficients.
        """
        weights = np.eye(self.q)
        for k in range(self.q):
            for l in range(self.q):
                weights[k][l] = 1 - abs(self.categ_vec[k] - self.categ_vec[l]
                                        ) / abs(self.xmax - self.xmin)
        return weights

    def ordinal(self):
        """ Function for computing the Ordinal Weights

        Ordinal weights of a matrix :math:`W \in R^{q \\times q}`, where
        :math:`q` is the number of `categories`, are defined for each cell
        :math:`w_{kl}, (k,l=1, \ldots, q)` by

        .. math::
            w_{kl} = \\frac{(\max(k, l) - \min(k, l) + 1) \cdot
                     (\max(k, l) - \min(k, l))}{2}

        Returns
        -------
        2D :math:`q \\times q` matrix
            A square matrix of ordinal weights to be used for calculating the
            weighted coefficients.
        """
        weights = np.eye(self.q)
        for k in range(self.q):
            for l in range(self.q):
                nkl = max(k, l) - min(k, l) + 1
                weights[k][l] = nkl * (nkl - 1) / 2
        weights = 1 - weights / np.max(weights)
        return weights

    def quadratic(self):
        """ Function for computing the Quadratic Weights

        Quadratic weights of a matrix :math:`W \in R^{q \\times q}`, where
        :math:`q` is the number of `categories`, are defined for each cell
        :math:`w_{kl}, (k,l=1, \ldots, q)` by

        .. math::
            w_{kl} = 1 - \\frac{(k - l)^2}{(q - 1)^2}

        Returns
        -------
        2D :math:`q \\times q` matrix
            A square matrix of quadratic weights to be used for calculating the
            weighted coefficients.
        """
        weights = np.eye(self.q)
        diff = self.xmax - self.xmin
        for k in range(self.q):
            for l in range(self.q):
                weights[k][l] = 1 - pow(
                    (self.categ_vec[k] - self.categ_vec[l]) / diff, 2)
        return weights

    def radical(self):
        """ Function for computing the Radical Weights

        Radical weights of a matrix :math:`W \in R^{q \\times q}`, where
        :math:`q` is the number of `categories`, are defined for each cell
        :math:`w_{kl}, (k,l=1, \ldots, q)` by

        .. math::
            w_{kl} = 1 - \\frac{|k - l|^{1/2}}{(\max(c) - \min(c))^{1/2}}

        where :math:`c \in R^q` is the vector of the `categories`.

        Returns
        -------
        2D :math:`q \\times q` matrix
            A square matrix of radical weights to be used for calculating the
            weighted coefficients.
        """
        weights = np.eye(self.q)
        for k in range(self.q):
            for l in range(self.q):
                weights[k][l] = 1 - np.sqrt(
                    abs(self.categ_vec[k] - self.categ_vec[l])) / np.sqrt(
                        abs(self.xmax - self.xmin))
        return weights

    def ratio(self):
        """ Function for computing the Ratio Weights

        Ratio weights of a matrix :math:`W \in R^{q \\times q}`, where
        :math:`q` is the number of `categories`, are defined for each cell
        :math:`w_{kl}, (k,l=1, \ldots, q)` by

        .. math::
            w_{kl} = 1 - (\\frac{k - l}{k + l})^2/(
                \\frac{\\max(c) - \\min(c)}{\\max(c) - \\min(c)})^2

        where :math:`c \in R^q` is the vector of the `categories`.

        Returns
        -------
        2D :math:`q \\times q` matrix
            A square matrix of ratio weights to be used for calculating the
            weighted coefficients.

        Raises
        ------
        ValueError
            In cases we have the 0 as a category. This will produce a division
            by zero error.
        """
        if 0 in self.categ_vec:
            raise ValueError(
                'You have 0 as a category. Please do not use'
                ' 0 as a category because it produce a'
                ' division by 0.')
        weights = np.eye(self.q)
        for k in range(self.q):
            for l in range(self.q):
                weights[k][l] = 1 - (
                    pow(
                        (self.categ_vec[k] - self.categ_vec[l]) /
                        (self.categ_vec[k] + self.categ_vec[l]), 2)) / pow(
                            (self.xmax - self.xmin) /
                            (self.xmax + self.xmin), 2)
        return weights



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['irrCAC']

package_data = \
{'': ['*']}

install_requires = \
['coverage>=6.2,<7.0',
 'pandas>=1.3.5,<2.0.0',
 'pytest>=6.2.5,<7.0.0',
 'scipy>=1.7.3,<2.0.0']

setup_kwargs = {
    'name': 'irrcac',
    'version': '0.2.0',
    'description': 'Degree of agreement among raters.',
    'long_description': 'Chance-corrected Agreement Coefficients\n=======================================\n\n.. image:: https://readthedocs.org/projects/irrcac/badge/?version=latest\n  :target: https://irrcac.readthedocs.io/en/latest/?badge=latest\n  :alt: Documentation Status\n\n\nThe **irrCAC** is a Python package that provides several functions for\ncalculating various chance-corrected agreement coefficients. This package\nclosely follows the general framework of inter-rater reliability assessment\npresented by Gwet (2014).\n\nThe functionality covers calculations for various chance-corrected agreement\ncoefficients (CAC) among 2 or more raters. Among the CAC coefficients covered\nare Cohen\'s kappa, Conger\'s kappa, Fleiss\' kappa, Brennan-Prediger coefficient,\nGwet\'s AC1/AC2 coefficients, and Krippendorff\'s alpha. Multiple sets of weights\nare proposed for computing weighted analyses.\n\nThe functions included in this package can handle 2 types of input data. Those\ntypes with the corresponding coefficients are in the following list:\n\n1. Contingency Table\n\n  1. Brennar-Prediger\n  2. Cohen\'s kappa\n  3. Gwet AC1/AC2\n  4. Krippendorff\'s Alpha\n  5. Percent Agreement\n  6. Schott\'s Pi\n\n2. Raw Data\n\n  1. Fleiss\' kappa\n  2. Gwet AC1/AC2\n\n.. note::\n   All of these statistical procedures are described in details in\n   Gwet, K.L. (2014,ISBN:978-0970806284):\n   "Handbook of Inter-Rater Reliability," 4th edition, Advanced Analytics, LLC.\n\n   This package is a port *(with permission)* to Python of the\n   `irrCAC <https://github.com/kgwet/irrCAC>`_ library for R by Gwet, K.L.\n\n.. important::\n   This is a **work in progress** and *does not* have (yet) the full\n   functionality found in the R library.\n\nInstallation\n------------\nTo install the package, run:\n\n.. code:: bash\n\n    pip install irrCAC\n\nDevelopers\n----------\nTo run the tests, install `poetry <https://python-poetry.org/>`_ and run:\n\n.. code:: bash\n\n    poetry install\n\nNext run:\n\n.. code:: bash\n\n    poetry run pytest\n\nThere is also a config file for `tox <https://tox.readthedocs.io/en/latest/>`_\nso you can automatically run the tests for various python versions like this:\n\n.. code:: bash\n\n    tox\n\nDocumentation\n-------------\nThe documentation of the project is available at the following page:\nhttp://irrcac.readthedocs.io/\n',
    'author': 'Aris Fergadis',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/afergadis/irrCAC',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
